#!/usr/bin/python
import argparse
import json
import logging
import os
from threading import Thread
import time

import api as api
import cli as cli
import pytools.experiments.job_generator as jobgen
from common.lx3.common import get_config_json
from mn.mn_launcher import mn_start, get_mininet_flags
from pytools.common.common import print_generic_exception
from pytools.ycsb.ycsb_runner import generate_sub_exp_folder_based_on_args
from utils.create_data_folder import create_folder
from utils.execute_ping_nodes_generate_output import \
    execute_ping_nodes_generate_output
from utils.plot_results_from_ping import plot_ping_results
from utils.ds_score_plots.fetch_score_and_write_to_csv import \
    fetch_data_from_log_and_write_to_csv
from utils.ds_score_plots.plot_time_vs_score import \
    plot_score_data

from utils.experiments.read_yscb_data_from_folder import \
    get_yscb_data_from_folder
from utils.experiments.write_results_to_csv import write_results_dict_to_csv
from utils.slack.slack_notification import generate_slack_notifications

logging.basicConfig(format="%(filename).7s.py (%(funcName).5s):[%(lineno)4d] %(asctime)s %(levelname)7s || %(message)s",
    datefmt='%H:%M:%S', level=logging.DEBUG)


def adjust_by_flags(FLAGS, conf):
    if FLAGS.setup_only:
        conf["cassandra"]["setup"]["cas_copy"] = True
        conf["cassandra"]["setup"]["cas_cfg"] = True
        conf["cassandra"]["setup"]["cas_rm_data"] = True
        conf["cassandra"]["setup"]["cas_run_experiment"] = False

    return conf

def run_multiconf_experiments(FLAGS, conf, unique_exp_folder):
    """
    Run multiconf experiments
    --> use latest conf.json version for it to work
    :param FLAGS:
    :param conf:
    """
    if FLAGS.repeat > 0:
        conf["cassandra"]["exp"]["repeat"] = FLAGS.repeat

    # Do initial setting up testbed here
    net = setup_testbed(FLAGS, conf, launch_system=True)
    # Load YCSB stuff to the keyspace
    api.experiment_setup_keyspace(FLAGS, conf, net=net)

    # Run experiments one after other
    print("######################################################")
    print("Muliconf multi thread experiment started")
    run_multiconf_threads_and_plot(
        FLAGS, conf, net=net, unique_exp_folder=unique_exp_folder
    )
    print("Muliconf multi thread experiment finished")
    print("######################################################")
    print("######################################################")
    print("Muliconf multi sleep time experiment started")
    run_multiconf_sleep_time(
        FLAGS, conf, net=net, unique_exp_folder=unique_exp_folder
    )
    print("Muliconf multi sleep time experiment finished")
    print("######################################################")
    print("######################################################")
    print("Muliconf multi jitter experiment started")
    run_multiconf_jitter(
        FLAGS, conf, net=net, unique_exp_folder=unique_exp_folder,
        distribution=FLAGS.distribution
    )
    print("Muliconf multi jitter experiment finished")
    print("######################################################")


def run_multiconf_experiments_with_every_snitch(FLAGS, conf):
    """
    Run multiconf experiments with every
    snitch version
    --> use latest conf.json version for it to work
    :param FLAGS:
    :param conf:
    """
    # Dynamic Snitch multiconf experiments
    conf["cassandra"]["config_items"]["dynamic_snitch"] = True
    conf["cassandra"]["config_items"]["fairgame_snitch"] = False
    conf["cassandra"]["config_items"]["omnipotent_snitch"] = False

    print("<-------------------------------------------------------->")
    print("Dynamic Snitch muliconf experiment started")
    dynamic_snitch_exp_folder =  create_experiment_folders("dynamic_snitch")
    run_multiconf_experiments(FLAGS, conf, dynamic_snitch_exp_folder)
    print("<-------------------------------------------------------->")

    # Omnipotent Snitch multiconf experiments
    conf["cassandra"]["config_items"]["dynamic_snitch"] = False
    conf["cassandra"]["config_items"]["fairgame_snitch"] = False
    conf["cassandra"]["config_items"]["omnipotent_snitch"] = True

    print("<-------------------------------------------------------->")
    print("Omnipotent Snitch muliconf experiment started")
    omnipotent_snitch_exp_folder =  create_experiment_folders("omnipotent_snitch")
    run_multiconf_experiments(FLAGS, conf, omnipotent_snitch_exp_folder)
    print("<-------------------------------------------------------->")

    # Fairgame Snitch multiconf experiments
    conf["cassandra"]["config_items"]["dynamic_snitch"] = False
    conf["cassandra"]["config_items"]["fairgame_snitch"] = True
    conf["cassandra"]["config_items"]["omnipotent_snitch"] = False

    print("<-------------------------------------------------------->")
    print("Fairgame Snitch muliconf experiment started")
    fairgame_snitch_exp_folder =  create_experiment_folders("fairgame_snitch")
    run_multiconf_experiments(FLAGS, conf, fairgame_snitch_exp_folder)
    print("<-------------------------------------------------------->")


def execute_args_workloads(FLAGS, conf, net):
    for w in FLAGS.workloads:
        try:
            wcfg = get_config_json(w)
            cfgs = jobgen.generate_jobs_configs_from_json(
                wcfg, FLAGS.properties, FLAGS.ranges)

            for c in cfgs:
                exp_name = "wl.%s_arg.%s_%s" % (
                    os.path.basename(w), c["job_short_id"], FLAGS.comment)
                exp_name = exp_name.replace("/", ".")
                if FLAGS.repeat > 0:
                    c["cassandra"]["exp"]["repeat"] = FLAGS.repeat

                api.experiment_run(FLAGS, c, net, exp_name)

        except:
            print_generic_exception()


def create_ping_plot_data_dirs(conf):
    """
    Create plot directories (out/plot_data/<>)
    :param conf:
    :return:
    """
    # Create output parent folder
    plot_data_folder = os.path.join(conf["env"]["out_exp_dir"], 'plot_data')
    create_folder(plot_data_folder)

    # Create child folders
    latency_data_folder = os.path.join(plot_data_folder, 'latency_data')
    create_folder(latency_data_folder)

    jitter_data_folder = os.path.join(plot_data_folder, 'jitter_data')
    create_folder(jitter_data_folder)

    ping_data_folder = os.path.join(plot_data_folder, 'ping_data')
    create_folder(ping_data_folder)

    score_data_folder = os.path.join(plot_data_folder, 'score_data')
    create_folder(score_data_folder)

    return


def set_default_test_bed_state(FLAGS, conf, net, round_repeat=False):
    # Standard defaults
    api.execute_const_update_jitter(
        FLAGS, conf, net, jitter_level=2, distribution="normal"
    )
    conf['workload_1']['threads'] = 256  # Default
    conf['workload_1']['p_recordcount'] = 10000
    conf['workload_1']['p_operationcount'] = 10000
    return conf


def run_multiconf_threads_and_plot(FLAGS, conf, net, unique_exp_folder):
    """
    Run experiment with varying YCSB thread loads and draw plots (1.5)
    :param FLAGS:
    :param conf:
    :return:
    """
    # We need the folders where these outputs are stored
    thread_count_results = []
    output_folder = None

    # This would set the defaults
    conf = set_default_test_bed_state(FLAGS, conf, net)

    for thread_count in conf['multiconf_num_threads']:
        result_for_thread = {
            'label': 'thread count',
            'key': thread_count,
            'result': None
        }
        conf['workload_1']['threads'] = thread_count
        output_folder = api.experiment_run(FLAGS, conf, net)
        print("**********************************************************")
        print(
            "Experiments finished for: {0} threads, and output at: {1}".format(
                thread_count, output_folder
            )
        )
        print("**********************************************************")

        # store intermediate results
        result_for_thread['result'] = get_yscb_data_from_folder(
            output_folder
        )
        thread_count_results.append(result_for_thread)

    write_results_dict_to_csv(
        results_dict=thread_count_results,
        location=unique_exp_folder,
        filename='Varying_thread_loads_levels.csv'
    )
    # All tests are finished at this point
    api.get_ycsb_plot_from_list(
        os.path.join(
            unique_exp_folder, 'Varying_thread_loads_levels.csv'
        ), where_to_save=unique_exp_folder,
        label="Varying thread load effects"
    )
    return


def run_multiconf_sleep_time(FLAGS, conf, net, unique_exp_folder):
    """
    Run experiment with varying sleep times and draw plots (1.5)
    :param FLAGS:
    :param conf:
    :return:
    """
    # We need the folders where these outputs are stored
    output_folder = None
    sleep_time_results = []

    # This would set the defaults
    conf = set_default_test_bed_state(FLAGS, conf, net)
    for sleep_time in conf['multiconf_sleep_time']:
        result_for_sleep_time = {
            'label': 'sleep time',
            'key': sleep_time,
            'result': None
        }
        # Execute latency change with different sleep times
        thread = Thread(
            target=api.execute_update_latencies,
            args=(FLAGS, conf, net, sleep_time)
        )
        thread.start()
        output_folder = api.experiment_run(FLAGS, conf, net)

        print("**********************************************************")
        print(
            "Experiments finished for: {0} sleep time, output at: {1}".format(
                sleep_time, output_folder
            )
        )
        print("**********************************************************")

        thread.join()

        # store intermediate results
        result_for_sleep_time['result'] = get_yscb_data_from_folder(
            output_folder
        )
        sleep_time_results.append(result_for_sleep_time)

    write_results_dict_to_csv(
        results_dict=sleep_time_results,
        location=unique_exp_folder,
        filename='Varying_latency_frequency_levels.csv'
    )

    # All tests are finished at this point
    api.get_ycsb_plot_from_list(
        os.path.join(
            unique_exp_folder, 'Varying_latency_frequency_levels.csv'
        ),
        where_to_save=unique_exp_folder,
        label="Varying sleep frequency changes effects"
    )


def run_constant_latency_and_jitter_exp(
    FLAGS, conf, net, jitter, latency, distribution='normal'
):
    """
    Run exp with worst case latency and jitter
    :param FLAGS:
    :param conf:
    :param net:
    :return:
    """

    api.execute_const_update_jitter(
        FLAGS, conf, net, jitter_level=jitter, distribution=distribution,
        latency=latency
    )
    return api.experiment_run(FLAGS, conf, net)


def run_multiconf_propagation_delta_fs(
    FLAGS, conf, net, jitter, latency, distribution='normal',
    unique_exp_folder=None
):
    prop_delta_min = conf['multiconf_propagation_delta']['min']
    prop_delta_max = conf['multiconf_propagation_delta']['max']
    delta_conf_results = []
    for delta in range(prop_delta_min, prop_delta_max):
        result_for_delta = {
            'label': 'Propogation delta',
            'key': delta,
            'result': None
        }
        api.execute_const_update_jitter(
            FLAGS, conf, net, jitter_level=jitter, distribution=distribution,
            latency=latency
        )
        conf['cassandra']['config_items']['fairgame_snitch_propagation_delta']\
            = delta

        output_folder = api.experiment_run(FLAGS, conf, net)
        print("**********************************************************")
        print(
            "Experiments finished for: {0} propogation delta, and output at: "
            "{1}".format(delta, output_folder)
        )
        print("**********************************************************")

        result_for_delta['result'] = get_yscb_data_from_folder(output_folder)
        delta_conf_results.append(result_for_delta)

    write_results_dict_to_csv(
        results_dict=delta_conf_results, location=unique_exp_folder,
        filename='Varying_propagation_delta_fs_score.csv'
    )

    # All tests are finished at this point
    api.get_ycsb_plot_from_list(
        os.path.join(
            unique_exp_folder, 'Varying_propagation_delta_fs_score.csv'
        ),
        where_to_save=unique_exp_folder,
        label="Varying propagation delta effects"
    )


def run_multiconf_jitter(FLAGS, conf, net, unique_exp_folder, distribution):
    """
    Run experiment with varying Jitter levels and draw plots (1.5)
    :param FLAGS:
    :param conf:
    :return:
    """
    # We need the folders where these outputs are stored
    output_folder = None
    jitter_level_results = []

    # This would set the defaults
    conf = set_default_test_bed_state(FLAGS, conf, net)
    for jitter_level in conf['multiconf_jitter_levels']:
        result_for_jitter = {
            'label': 'jitter level',
            'key': jitter_level,
            'result': None
        }
        # Modify jitter level here with value
        api.execute_const_update_jitter(
            FLAGS, conf, net, jitter_level, distribution=distribution
        )

        output_folder = api.experiment_run(FLAGS, conf, net)

        print("**********************************************************")
        print(
            "Experiments finished for: {0} jitter, and output at: {1}".format(
                jitter_level, output_folder
            )
        )
        print("**********************************************************")

        # store intermediate results
        result_for_jitter['result'] = get_yscb_data_from_folder(
            output_folder
        )
        jitter_level_results.append(result_for_jitter)

    write_results_dict_to_csv(
        results_dict=jitter_level_results,
        location=unique_exp_folder,
        filename='Varying_jitter_levels.csv'
    )

    # All tests are finished at this point
    api.get_ycsb_plot_from_list(
        os.path.join(unique_exp_folder, 'Varying_jitter_levels.csv'),
        where_to_save=unique_exp_folder,
        label="Varying jitter level effects"
    )


def create_experiment_folders(unique_name=''):
    create_folder(
        os.path.join(conf["env"]["out_exp_dir"], 'exp_out'),
        remove_if_exists=False
    )
    # Create a unique folder for these runs, and save things there
    if unique_name:
        unique_exp_folder_name = \
            'exp-{0}-{1}-results'.format(unique_name, int(time.time()))
    else:
        unique_exp_folder_name = 'exp-{0}-results'.format(int(time.time()))

    create_folder_path = os.path.join(
        conf["env"]["out_exp_dir"], 'exp_out', unique_exp_folder_name
    )
    create_folder(create_folder_path, remove_if_exists=False)
    return create_folder_path


def setup_testbed(FLAGS, conf, launch_system=False):
    """
    Basic setting up of testbed, usually required for every experiment
    :param FLAGS:
    :param conf:
    :param launch_system:
    :return net mininet instance for future use
    """
    # 0. CLEANUP
    api.experiment_cleanup(FLAGS, conf)

    # 1. PREPARE
    api.experiment_prepare(FLAGS, conf)

    # 2. SETUP MININET TOPO
    net = mn_start(FLAGS, conf)

    # 3. LAUNCH SYSTEMS
    if launch_system:
        api.experiment_launch_systems(FLAGS, conf, net)

    return net


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        "lxIII",  add_help=False,
        parents=[get_mininet_flags(), jobgen.get_generators_argparser()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        '---Launcher Flags---',
        help='-----------------------------------------------------',
        action='store_true')
    # parser.add_argument("--cfg", default="./experiments/sample/sample.json",
    #   help="A .json configuration file for the experiment")
    parser.add_argument(
        "--recreate_keyspace", action="store_true",
        help="""YCSB key space will be created from scratch according to
        config. YCSB will be executed to populate the keyspace afterwards""")
    # Pass this option to repeat experiments with multiple configurations
    parser.add_argument(
        "--multiconf", action="store_true",
        help="""Causes multiple experiments with different configurations""")
    parser.add_argument(
        "--multiconf_propagation_delta", action="store_true",
        help='Causes multiple experiments with different delta configurations')
    parser.add_argument(
        "--every_snitch", action="store_true",
        help='Causes multiple experiments with every snitch version')
    parser.add_argument(
        "--vary_threads", action="store_true",
        help="""Causes multiple experiments with different YCSB thread 
        configurations""")
    parser.add_argument(
        "--vary_latency", action="store_true",
        help="""Causes multiple experiments with different latency 
        configurations""")
    parser.add_argument(
        "--vary_jitter", action="store_true",
        help="""Causes multiple experiments with different jitter 
        configurations""")
    parser.add_argument(
        "--worst_case", action="store_true",
        help='Simulate a worst case scenario with maximum latency and jitter'
    )
    parser.add_argument(
        "--latency", type=int, default=0,
        help='Latency value for worst_case'
    )
    parser.add_argument(
        "--jitter", type=int, default = 0,
        help='Jitter value for worst_case'
    )
    parser.add_argument(
        "--distribution", default='normal',
        help='Jitter distribution name'
    )
    parser.add_argument(
        "--setup_only", action="store_true",
        help="""Does not runs experiment, but only sets up everything from 
        scratch. Option is propagated by overwriting certain fields in the 
        config json.""")
    parser.add_argument(
        "--workloads", nargs='*', default=[],
        help="List of workloads to execute after initial setup")
    parser.add_argument(
        "--repeat", type=int, default=0,
        help="Overwrites the number of times each experiment will be repeated")
    parser.add_argument(
        "--test", action="store_true",
        help="Executes a single dynamic latency/jitter experiment with "
             "ping tests."
    )
    parser.add_argument(
        "--vary_latency_score", action="store_true",
        help="""Causes multiple experiments with different latency 
        configurations and collect score data""")
    FLAGS = parser.parse_args()
    # Notify our chatbot about this operation
    generate_slack_notifications(message="Test run started")

    conf = json.load(FLAGS.cfg_json)
    conf = adjust_by_flags(FLAGS, conf)
    create_ping_plot_data_dirs(conf=conf)

    if FLAGS.multiconf:
        unique_exp_folder = create_experiment_folders()
        # This is now in a separate function for simpler reuse
        run_multiconf_experiments(FLAGS, conf, unique_exp_folder)

    elif FLAGS.every_snitch:
        run_multiconf_experiments_with_every_snitch(FLAGS, conf)

    elif FLAGS.setup_only:
        # Setup only is not going to have Cassandra instances therefore we can
        # ignore notify_cassandra
        conf['notify_cassandra'] = False
        net = setup_testbed(FLAGS, conf, launch_system=False)
    elif FLAGS.test: # Ping tests has a dedicated flag
        if conf['notify_cassandra']:
            raise Exception(
                "Error: notify_cassandra is set to true while there are no "
                "Cassandra instances"
            )
        net = setup_testbed(FLAGS, conf, launch_system=False)
        # start pinging and saving data here (so that we dont have to do it
        # again later)
        execute_ping_nodes_generate_output(conf=conf, net=net)

        # Here we should execute further changes to testbed and plot
        api.execute_update_latencies(
            FLAGS, conf, net, frequency=10, round_repeat=False
        )
        plot_ping_results(
            conf,
            target_folder=os.path.join(
                conf["env"]["out_exp_dir"], 'plot_data', 'latency_data'
            ),
            where_to_save=os.path.join(
                conf["env"]["out_exp_dir"], 'plot_data', 'latency_data'
            ),
            title='Round-Trip-Time vs Time at varying Latency'
        )
        plot_ping_results(
            conf,
            target_folder=os.path.join(
                conf["env"]["out_exp_dir"], 'plot_data', 'jitter_data'
            ),
            where_to_save=os.path.join(
                conf["env"]["out_exp_dir"], 'plot_data', 'score_data'
            ),
            title='Round-Trip-Time vs Time at varying Jitter'
        )
    elif FLAGS.worst_case:
        # If this is made - make super bad latency and jitter configs
        net = setup_testbed(FLAGS, conf, launch_system=True)
        api.experiment_setup_keyspace(FLAGS, conf, net=net)
        print("#############################################################")
        print("Constant worst case test started")
        # execute_ping_nodes_generate_output(conf=conf, net=net)
        run_constant_latency_and_jitter_exp(
            FLAGS, conf, net=net, latency=FLAGS.latency, jitter=FLAGS.jitter,
            distribution=FLAGS.distribution
        )
        print("Constant worst case test finished")

    elif FLAGS.multiconf_propagation_delta:
        unique_exp_folder = create_experiment_folders()
        # If this is made - make super bad latency and jitter configs
        net = setup_testbed(FLAGS, conf, launch_system=True)
        api.experiment_setup_keyspace(FLAGS, conf, net=net)
        print("#############################################################")
        print("Multiple propogation delta test started")
        run_multiconf_propagation_delta_fs(
            FLAGS, conf, net=net, jitter=FLAGS.jitter, latency=FLAGS.latency,
            distribution=FLAGS.distribution,
            unique_exp_folder=unique_exp_folder
        )
        print('Multiple propogation delta test finished')
        print("#############################################################")
    else: # If no flag has been provided, we give CLI back to the user
        net = setup_testbed(FLAGS, conf, launch_system=True)
        # Load YCSB stuff to the keyspace
        api.experiment_setup_keyspace(FLAGS, conf, net=net)
        cli.run_cli(FLAGS, conf, net)
